(defpackage :text.csv (:use :cl) (:nicknames :csv) (:export #:parse-csv-file #:parse-csv))
(in-package :text.csv)

(defun parse-csv-file (file &key (external-format #+clisp system::*default-file-encoding*
						  #+allegro excl:*default-external-format*
						  #+sbcl :utf-8)
		                 (separator #\,) (quote #\")
		                 (buffer-size 8192) (field-size 1024) (adjustable nil))
  (if (null (probe-file file))
      (error "can't open file")
      (with-open-file (stream file :direction :input :external-format external-format
                              #+clisp :buffered #+clisp t
                              #+allegro :mapped #+allegro t)
        (parse-csv stream :separator separator :quote quote
                          :buffer-size buffer-size :field-size field-size :adjustable adjustable))))

(defun parse-csv (stream &key (separator #\,) (quote #\") (buffer-size 8192) (field-size 1024) (adjustable nil))
  (declare (optimize (speed 3) (space 0) (debug 0) (safety 0)))
  (declare (type character separator quote))
  (declare (type fixnum buffer-size field-size))
  (let* ((buffer      (make-string buffer-size :initial-element #\Null))
         (field       (make-array field-size :adjustable adjustable :element-type 'character :initial-element #\Null :fill-pointer nil))
         (offset      0)
         (space-count 0)
         (row         nil)
         (row-ptr     nil)
         (result      nil)
         (result-ptr  nil)
         (state       :START))
    (declare (type fixnum offset space-count))
    (labels ((extend-field ()
               (setf field-size (the fixnum (* 2 (the fixnum field-size)))
                     field      (adjust-array field field-size)))
             (cell-input (char)
	       (declare (type character char))
               (when (>= (the fixnum offset) (the fixnum field-size))
                 (if (null adjustable)
                     (error "too large field size")
                     (extend-field)))
               (setf (char field offset) char)
	       (incf offset))
             (cell-finish ()
               (let ((cell (subseq field 0 (if (> (the fixnum space-count) 0)
                                               (the fixnum (- (the fixnum offset) (the fixnum space-count)))
                                               offset))))
                 (setf offset 0
                       space-count 0)
                 (if (null row)
                     (setf row (cons cell nil)
                           row-ptr row)
                     (setf (cdr row-ptr) (cons cell nil)
                           row-ptr       (cdr row-ptr)))))
             (row-finish ()
               (when (> (the fixnum offset) 0)
                 (cell-finish))
               (if (null result)
                   (setf result (cons row nil)
                         result-ptr result)
                   (setf (cdr result-ptr) (cons row nil)
                         result-ptr       (cdr result-ptr)))
               (setf row nil))
             (parse (buffer size)
               (loop for i of-type fixnum from 0 below size
                     for c of-type character across buffer do
                     (tagbody
                        (when (and (eq state :QUOTE-ESCAPSE)
                                   (char/= c quote))
                          (setf state :START))
                        (when (eq state :START)
                          (cond ((or (char= c #\Space) (char= c #\Tab))
                                 (go :CONTINUE))
                                ((or (char= c #\Newline) (char= c #\Linefeed))
                                 (row-finish)
                                 (go :CONTINUE))
                                ((char= c separator)
                                 (cell-finish)
                                 (go :CONTINUE)) 
                                ((char= c quote)
                                 (setf state :QUOTE)
                                 (go :CONTINUE))
                                (t
                                 (setf state :NORMAL))))
                        ;; processing field
                        (case state
                          (:QUOTE
                           (cond ((char= c quote)
                                  (setf state :QUOTE-ESCAPSE)
                                  (go :CONTINUE))
                                 (t
                                  (cell-input c))))
                          (:NORMAL
                           (cond ((char= c separator)
                                  (cell-finish)
                                  (setf state :START))
                                 ((or (char= c #\Newline) (char= c #\Linefeed))
                                  (row-finish)
                                  (setf state :START))
                                 ((or (char= c #\Space) (char= c #\Tab) (char= c #\Return))
                                  (cell-input c)
                                  (setf space-count (the fixnum (+ 1 (the fixnum space-count)))))
                                 (t
                                  (cell-input c)
                                  (setf space-count 0))))
                          (:QUOTE-ESCAPSE
                           (cond ((char= c quote)
                                  (cell-input c)
                                  (setf state :QUOTE)))))
                      :CONTINUE)))
             (finish ()
               (when (> (the fixnum offset) 0)
                 (row-finish))
               result))
      (declare (inline extend-field cell-input cell-finish row-finish parse finish))
      (loop for size = (read-sequence buffer stream)
            until (= size 0)
            do (parse buffer size)
            finally (return (finish))))))

#||
		  
(defpackage :text.csv (:use :cl) (:nicknames :text.csv.gauche) (:export #:make-csv-reader #:make-csv-writer))
(in-package :text.csv.gauche)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Gauche Compatible Layer
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun string-null? (string)
  (string= string ""))

(defun string-take (string nchars)
  (subseq string 0 nchars))

(defun string-drop (string nchars)
  (subseq string nchars))

(defun string-concatenate (lst)
  (apply #'concatenate (cons 'string lst)))

(defun string-concatenate-reverse (lst)
  (apply #'concatenate (cons 'string (reverse lst))))

(defun string-prefix? (prefix string)
  (when (>= (length string) (length prefix))
    (loop for c1 across prefix
          for c2 across string
          always (char= c1 c2))))

(defun string-index (string chars)
  (loop for c in chars thereis (position c string)))

(defun string-scan (string item &optional (return :index))
  (multiple-value-bind (pos offset)
      (typecase item
	(string (values (search item string) (length item)))
	(character (values (position item string) 1)))
    (if (null pos)
	(case return
	  ((:index :before :after)  nil)
	  ((:before* :after* :both) nil)
	  (t                        (error "invalid return option")))
	(case return
	  (:index   pos)
	  (:before  (subseq string 0 pos))
	  (:after   (subseq string (+ pos offset)))
	  (:before* (values (subseq string 0 pos) (subseq string pos)))
	  (:after*  (values (subseq string 0 (+ pos offset)) (subseq string (+ pos offset))))
	  (:both    (values (subseq string 0 pos) (subseq string (+ pos offset))))
	  (t        (error "invalid return option"))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; CSV Reader/Writer
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun make-csv-reader (&key (separator #\,) (quote #\") (newline '(#\Newline))
			&aux (quote-string (coerce (list quote) 'string))
			     (newline-string (coerce newline 'string)))
  (lambda (stream)
    (labels ((start (line fields)
	       (if (eql line :eof)
		   (nreverse fields)
		   (let ((next (string-trim " " line)))
		     (if (string-prefix? quote-string next)
			 (quoted (string-drop next 1) fields ())
			 (noquote next fields)))))
	     (noquote (line fields &aux (i (position separator line)))
	       (if i
		   (start (string-drop line (+ i 1))
			  (cons (string-right-trim " " (string-take line i)) fields))
		   (nreverse (cons (string-right-trim " " line) fields))))
	     (quoted (line fields partial)
	       (cond ((eql line :eof) (error "unterminated quoted field"))
		     ((string-null? line)
		      (quoted (read-line stream nil :eof) fields (cons newline-string partial)))
		     (t
		      (multiple-value-bind (this next)
			  (string-scan line quote :both)
			(if this
			    (if (string-prefix? quote-string next)
				(quoted (string-drop next 1) fields
					(list* quote-string this partial))
				(let ((next-next (string-scan next separator :after))
				      (f (string-concatenate-reverse (cons this partial))))
				  (if next-next
				      (start next-next (cons f fields))
				      (nreverse (cons f fields)))))
			    (quoted (read-line stream nil :eof) fields (list* newline-string line partial))))))))
      (let ((line (read-line stream nil :eof)))
        (if (eql line :eof)
	    line
	    (start line ()))))))

(defun make-csv-writer (&key (separator #\,) (quote #\") (newline '(#\Newline))
			&aux (quote-string (coerce (list quote) 'string))
			     (newline-string (coerce newline 'string)))
  (let* ((quote-escape (concatenate 'string quote-string quote-string))
	 (separator-chars (if (stringp separator)
                              (coerce separator 'list)
                              (list separator)))
	 (special-chars (list* quote #\Space #\Newline #\Return separator-chars)))
    (lambda (stream fields)
      (let ((buffer (with-output-to-string (s)
		      (flet ((write-a-field (field)
			       (if (string-index field special-chars)
				   (format s "~A~A~A"
					   quote-string
					   (with-output-to-string (tmp)
					     (loop for c across field do
						   (if (char= c quote)
						       (write-string quote-escape tmp)
						       (write-char c tmp))))
					   quote-string)
				   (write-string field s))))
			(when fields
			  (write-a-field (car fields))
			  (loop for field in (cdr fields) do
			       (write-char separator s)
			       (write-a-field field)))
			(write-string newline-string s)))))
	(write-sequence buffer stream)))))


(defun test ()
  (let ((fun (make-csv-reader)))
    (with-input-from-string (s "\"11\",\"22\",\"33\",\"4
4\",\"\"\"555555
6666
7777\"")
      (let ((csv (loop for row = (funcall fun s)
		    until (eql row :eof)
		    collect row)))
	(values csv (with-output-to-string (s)
		      (loop with writer = (make-csv-writer)
			    for row in csv
			    do (funcall writer s row))))))))

||#
